//
// Copyright 1997 Steve Passe
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1997 Revision
// AUTHOR:     Steve Passe <smp@csn.net>
//

static char rcsid[] = "$Id: xtvremote.cc,v 1.18 1997/03/22 22:16:21 smp Exp smp $";

#define DEBUG_NOT


#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>

#include <sys/ioctl.h>
#include <machine/ioctl_meteor.h>
#include <machine/ioctl_bt848.h>	/* extensions to ioctl_meteor.h */

// Xlib headers
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>

// Standard Xt headers
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

// FWF Widget Set headers
#include <Xfwf/Board.h>
#include <Xfwf/Label.h>
#include <Xfwf/Button.h>
#include <Xfwf/PullDown.h>
#include <Xfwf/OptButton.h>
#include <Xfwf/TextMenu.h>
#include <Xfwf/RowCol.h>


class Action {

 public:

    Action() {}
    virtual ~Action() {}
    virtual int action( Widget w ) { return -1; }
};

class System : public Action {

 public:

    System( const char* args );
    virtual ~System() { delete _args; }
    virtual int action( Widget w );

 private:

    char*	_args;

};

typedef		int(*func_t)(Widget w, char*);

class Function : public Action {

 public:

    Function( int (*function)(Widget, char*), const char* args );
    Function( int (*function)(Widget) );
    virtual ~Function() { if ( _args ) delete _args; }
    virtual int action( Widget w );

// private:

    int		(*_func)(Widget, char*);
    char*	_args;

};

#define USE_ACCELERATORS_NOT

#define OK			0
#define ERROR			-1

#define SKIPWS(P)		while( isspace((*P)) ) ++P

// path for device specific resource files
#define RES_BASE		BASE "/app-defaults/"

#define BUTTON_RES_TMPL		"%s_b%02d"
#define MESSAGE_RES_TMPL	"%s_m%02d"

// maximum number of buttons per device
#define NUM_BUTTONS		100

// maximum number of devices
#define NUM_DEVICES		16
#define DEFAULT_DEVICE		"1"

// instance & class names for X resource access
const static char* appName = "xtvremote";
const static char* appClass = "Xtvremote";

// 
char* displayFontName = "*-helvetica-bold-r-*-140-*";


//
// application resource setup
//
#define	XtNxDevice	"xDevice"
#define	XtCXDevice	"XDevice"
#define	XtNxDeviceList	"xDeviceList"
#define	XtCXDeviceList	"XDeviceList"
#define	XtNxFavorites	"xFavorites"
#define	XtCXFavorites	"XFavorites"
#define	XtNxMixer	"xMixer"
#define	XtCXMixer	"XMixer"
#define	XtNxDisplayFont	"xDisplayFont"
#define	XtCXDisplayFont	"xDisplayFont"

typedef struct {
    char*	xDevice;
    char*	xDeviceList;
    char*	xFavorites;
    char*	xMixer;
    char*	xDisplayFont;
} appResources_t, *appResourcesPtr;

static appResources_t	appResources;

static XtResource	resources[] = {
#define offset(field)		XtOffset(appResourcesPtr, field)

	{ XtNxDevice, XtCXDevice, XtRString, sizeof(char*),
		offset(xDevice), XtRImmediate, (XtPointer)DEFAULT_DEVICE },

	{ XtNxDeviceList, XtCXDeviceList, XtRString, sizeof(char*),
		offset(xDeviceList), XtRImmediate, (XtPointer)NULL },

	{ XtNxFavorites, XtCXFavorites, XtRString, sizeof(char*),
		offset(xFavorites), XtRImmediate, (XtPointer)NULL },

	{ XtNxMixer, XtCXMixer, XtRString, sizeof(char*),
		offset(xMixer), XtRImmediate, (XtPointer)NULL },

	{ XtNxDisplayFont, XtCXDisplayFont, XtRString, sizeof(char*),
		offset(xDisplayFont), XtRImmediate, (XtPointer)NULL },
#undef offset
};

static XrmOptionDescRec options[] = {
	{ "-d",	".xDevice",	XrmoptionSepArg, (caddr_t) DEFAULT_DEVICE },
	{ "-l",	".xDeviceList",	XrmoptionSepArg, (caddr_t) NULL },
	{ "-l",	".xFavorites",	XrmoptionSepArg, (caddr_t) NULL },
	{ "-l",	".xMixer",	XrmoptionSepArg, (caddr_t) NULL },
	{ "-l",	".xDisplayFont",	XrmoptionSepArg, (caddr_t) NULL },
};


//
// global variables
//
static int	xDevice = 0;

//
// data for pulldown menu
//
typedef struct {
    int		index;		// index in array
    char*	label;		// ascii label
    char*	resFile;	// resource file
} deviceTag;


// label to x device associations for device menu
deviceTag deviceList[ NUM_DEVICES + 1 ] = {
    { -1, NULL,	NULL }
};

typedef struct {
    WidgetList	w;		// an array of Widget
    Action**	a;		// an array of Action
    int		c;		// count of each array
    int		x;		// width
    int		y;		// heigth
} device;

device		devices[ NUM_DEVICES ];

#define MAX_CHANNEL	125
static u_char	favoriteChannels[ MAX_CHANNEL ];
static int	favoriteCount = 0;
static int	currentChannel = 0;
static int	previousChannel = 0;
static int	inputChannel = 0;
static int	inputDigits = 0;

static int	currentHue = 0;
static int	currentBright = 0;
static int	currentContrast = 0;
static int	currentChroma = 0;
static int	currentSatU = 0;
static int	currentSatV = 0;
static int	currentSource = 0;
static int	currentTunerType = 0;

static int	currentMute = 0;
static int	currentVolume = 50;
static char*	mixerCommand = NULL;

// button data
typedef struct {
    Widget	w;
    Action*	a;
} button;

button		buttons[ NUM_BUTTONS ];

Widget		topLevel;

// connection stuff
static Action*	addMessage( char* args );

// callback functions
static void optionsCB( Widget, XtPointer, XtPointer );
static void deviceCB( Widget, XtPointer, XtPointer );
static void buttonCB( Widget w, XtPointer clientData, XtPointer callData );

// other local functions

static int	buildDeviceList( appResourcesPtr appRes, deviceTag* list );
static int	buildFavorites( appResourcesPtr appRes, u_char* list );
static int	loadDeviceResources( Widget w, deviceTag* list );
static int	buildDeviceMenu( Widget menu, deviceTag* list, int& index );
static void	setActiveMenu( Widget menu, deviceTag* list );

static int	buildButtonBoard( Widget ButtonBoard, deviceTag* deviceList );
static int	createButton( const char* deviceName, Widget parent,
			      int number, XtCallbackProc callBack );

// the hook to the STB card
static int	video;

// funk
static Widget*	widgetPointer = NULL;
static Widget	enterInputWidget = NULL;

static Widget	chnlDisplay = NULL;
static Widget	hueDisplay = NULL;
static Widget	brightDisplay = NULL;
static Widget	contrastDisplay = NULL;
static Widget	chromaDisplay = NULL;
static Widget	satUDisplay = NULL;
static Widget	satVDisplay = NULL;

static Widget	sourceTglW = NULL;
static Widget	tunerTglW = NULL;

static int	setChannel( Widget w, char* arg );
static int	getChannel( Widget w );
static int	updateChannel( int channel );
static int	getHue( Widget w );
static int	bumpHue( Widget w, char* arg );
static int	getBright( Widget w );
static int	bumpBright( Widget w, char* arg );
static int	getContrast( Widget w );
static int	bumpContrast( Widget w, char* arg );
static int	getChroma( Widget w );
static int	bumpChroma( Widget w, char* arg );
static int	getSatU( Widget w );
static int	bumpSatU( Widget w, char* arg );
static int	getSatV( Widget w );
static int	bumpSatV( Widget w, char* arg );
static int	bumpVolume( Widget w, char* arg );
static void	displayHue( int value );
static void	displayBright( int value );
static void	displayContrast( int value );
static void	displayChroma( int value );
static void	displaySatU( int value );
static void	displaySatV( int value );
static void	displayColor( void );
static void	displaySource( void );

static int	getSource( void );
static int	getTunerType( void );

#define XTVR_SHELLWIN_MASK (StructureNotifyMask | EnterWindowMask)
static void	XTVRShellWinEventHdlr( Widget, XtPointer, XEvent*, Boolean * );

struct {
    int		source;
    char*	name;
} sources[] = {
  { METEOR_INPUT_DEV0, "Video" },
  { METEOR_INPUT_DEV1, "Tuner" },
  { METEOR_INPUT_DEV2, "S-VHS" }
};

struct {
    int		tuner;
    char*	name;
} tunerTypes[] = {
  { 1, "Broadcast" },
  { 2, "IRC Cable" },
  { 3, "HRC Cable" }
};

//
// constructor, store args for action
//
System::System( const char* args )
{
    // keep a copy of the args
    _args = new char[ strlen( args ) + 1 ];
    strcpy( _args, args );

    return;
}


//
// do local system() call
//
int
System::action( Widget w )
{

    if ( _args )
    {
	system( _args );
    }

    return OK;
}


//
// constructor, store args for action
//
Function::Function( int (*function)(Widget, char*), const char* args )
{
    // keep a copy of the args
    _args = new char[ strlen( args ) + 1 ];
    strcpy( _args, args );

    _func = (func_t)function;

    return;
}


//
// constructor, store args for action
//
Function::Function( int (*function)(Widget) )
{
    _args = NULL;

    _func = (func_t)function;

    return;
}


//
// do local Function() call
//
int
Function::action( Widget w )
{
    return _func( w, _args );
}


//
// 
//
int
main( int argc, char **argv )
{
    XtAppContext	appContext;
    Widget		mainBox;
    Widget		mainMenu;
    Widget		optionsMenu;
    Widget		deviceMenu;
    Widget		ButtonBoard;

    // create top level shell
    topLevel = XtVaAppInitialize( &appContext, appClass,
				  options, XtNumber(options),
				  &argc, argv, NULL, NULL );

    // process remaining commandline args
    XtGetApplicationResources( topLevel,
			       &appResources, resources, XtNumber(resources),
			       0, 0 );

    if ((video = open("/dev/bktr0", O_RDONLY)) < 0) {
        fprintf(stderr, "open failed: %s\n", strerror(errno));
	exit( 1 );
    }


    // desired default device
    xDevice = atoi( appResources.xDevice ) - 1;

    if ( buildDeviceList( &appResources, deviceList ) == ERROR )
    {
	cerr << "BAD device list, missing Xremote app-default?" << endl;
	exit( 1 );
    }

    // 
    if ( buildFavorites( &appResources, favoriteChannels ) == ERROR )
    {
	cerr << "BAD favorites list, missing Xremote app-default?" << endl;
	exit( 1 );
    }

    // load device specific rdbs
    loadDeviceResources( topLevel, deviceList );

    // the top level container
    mainBox = XtVaCreateManagedWidget( "mainBox", xfwfBoardWidgetClass,
				       topLevel, NULL );

    //
    // build the menu bar
    //
    mainMenu = XtVaCreateManagedWidget( "menu", xfwfBoardWidgetClass,
					mainBox, NULL );

    optionsMenu = XtVaCreateManagedWidget( "oMenu", xfwfPullDownWidgetClass,
					   mainMenu, NULL );

    XtAddCallback( optionsMenu, XtNactivate, optionsCB, NULL );

    // create the device menu
    deviceMenu = XtVaCreateManagedWidget( "dMenu", xfwfOptionButtonWidgetClass,
					  mainMenu, NULL );

    XtAddCallback( deviceMenu, XtNactivate, deviceCB, NULL);

    // build device menu
    if ( buildDeviceMenu( deviceMenu, deviceList, xDevice ) == ERROR )
    {
	cerr << "BAD menu data" << endl;
	exit( 1 );
    }

#if defined( USE_ACCELERATORS )
    XtInstallAllAccelerators( topLevel, topLevel );
    XtInstallAllAccelerators( optionsMenu, topLevel );
#endif  // USE_ACCELERATORS

    //
    // button container
    //
    ButtonBoard = XtVaCreateManagedWidget( "keys", xfwfBoardWidgetClass,
					   mainBox, NULL );

    // create the button arrays
    cerr << "Building buttons, be patient, this takes time...";
    buildButtonBoard( ButtonBoard, deviceList );
    cerr << "done!" << endl;

    // activate only those devices that built correctly
    setActiveMenu( deviceMenu, deviceList );

    // change size
    Arg args[ 2 ];
    int i = 0;
    XtSetArg( args[i], XtNwidth, devices[ xDevice ].x );  i++;
    XtSetArg( args[i], XtNheight, devices[ xDevice ].y );  i++;
    XtSetValues( topLevel, args, i );

    // load a larger font for the channel display
    XFontStruct* fontInfo = NULL;
    if ( chnlDisplay )
    {
        Display* display = XtDisplay( topLevel );

        if ( appResources.xDisplayFont )
        {
            fontInfo = XLoadQueryFont( display, appResources.xDisplayFont );
            if ( fontInfo == NULL )
            {
                fprintf( stderr, "couldn't load font: '%s'\n",
                         appResources.xDisplayFont );
            }
        }

        if ( fontInfo == NULL )
        {
            fontInfo = XLoadQueryFont( display, displayFontName );
            if ( fontInfo == NULL )
            {
                fprintf( stderr, "couldn't load font: '%s'\n",
                         displayFontName );
            }
        }

        if ( fontInfo )
        {
            XtVaSetValues( chnlDisplay, XtNfont, fontInfo, NULL );
        }
    }

    if ( hueDisplay && fontInfo )
    {
            XtVaSetValues( hueDisplay, XtNfont, fontInfo, NULL );
    }

    if ( brightDisplay && fontInfo )
    {
            XtVaSetValues( brightDisplay, XtNfont, fontInfo, NULL );
    }

    if ( contrastDisplay && fontInfo )
    {
            XtVaSetValues( contrastDisplay, XtNfont, fontInfo, NULL );
    }

    if ( chromaDisplay && fontInfo )
    {
            XtVaSetValues( chromaDisplay, XtNfont, fontInfo, NULL );
    }

    if ( satUDisplay && fontInfo )
    {
            XtVaSetValues( satUDisplay, XtNfont, fontInfo, NULL );
    }

    if ( satVDisplay && fontInfo )
    {
            XtVaSetValues( satVDisplay, XtNfont, fontInfo, NULL );
    }

    // attempt to get the current channel selection
    getChannel( chnlDisplay );

    /* deal with freshly started system */
    if ( currentChannel == 0 )
    {
        char s[ 10 ];
        if ( favoriteCount )
        {
            sprintf( s, "%d", favoriteChannels[ 0 ] );
        }
        else
        {
            strcpy( s, "2" );
        }
        setChannel( chnlDisplay, s );
    }

    displaySource();
    displayColor();

    // the audio control
    if ( appResources.xMixer != NULL )
    {
        // get string into a local buffer
        mixerCommand = new char[ strlen(appResources.xMixer) + 1 ];
        strcpy( mixerCommand, appResources.xMixer );
        bumpVolume( (Widget)NULL, "0" );
    }

    // 
    XtManageChildren( devices[ xDevice ].w, devices[ xDevice ].c );

    // track changes to tv state by other clients
    XtAddEventHandler( topLevel, XTVR_SHELLWIN_MASK, False,
                       XTVRShellWinEventHdlr, NULL );

    // make it all happen
    XtRealizeWidget( topLevel );

    // ...if only this was done with threads...
    XtAppMainLoop( appContext );
}


//
// co-ordinate updates to controls with other clients
// borrowed from fxtv
//
void
XTVRShellWinEventHdlr( Widget w, XtPointer clientData,
                       XEvent *ev, Boolean *continue_dispatch )
{
    int	 x;

    switch ( ev->type )
    {
    case UnmapNotify:
        /// ???
        break;

    case ConfigureNotify:
        /// ???
        break;

    case EnterNotify:
        //  Resync menu options and tools with driver defaults
        if ( ioctl( video, TVTUNER_GETCHNL, &x ) < 0 )
          fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        if ( x != currentChannel )
          updateChannel( x );

        displaySource();
        displayColor();

        break;
    }
}


//
// 
//
void
bomb( char *s )
{
        perror( s );
        exit( 1 );
}

//
// update the current channel
//
static void
displaySource( void )
{
    char label[ 128 ];

    // get the current picture values
    if ( (currentSource = getSource()) < 0 )
      bomb( "couldn't access source reg" );
    XtVaSetValues( sourceTglW, XtNlabel,
                   sources[ currentSource-1 ].name, NULL );

    if ( (currentTunerType = getTunerType()) < 0 )
        bomb( "couldn't access source reg" );
    XtVaSetValues( tunerTglW, XtNlabel,
                   tunerTypes[ currentTunerType-1 ].name, NULL );
}


//
// update the current picture values
//
static void
displayColor( void )
{
    char label[ 128 ];

    if ( (currentHue = getHue( hueDisplay )) < 0 )
        bomb( "couldn't access hue reg" );
    displayHue( currentHue );

    if ( (currentBright = getBright( brightDisplay )) < 0 )
        bomb( "couldn't access bright reg" );
    displayBright( currentBright );

    if ( (currentContrast = getContrast( contrastDisplay )) < 0 )
        bomb( "couldn't access contrast reg" );
    displayContrast( currentContrast );

    if ( (currentChroma = getChroma( chromaDisplay )) < 0 )
        bomb( "couldn't access chroma reg" );
    displayChroma( currentChroma );

    if ( (currentSatU = getSatU( satUDisplay )) < 0 )
        bomb( "couldn't access satU reg" );
    displaySatU( currentSatU );

    if ( (currentSatV = getSatV( satVDisplay )) < 0 )
        bomb( "couldn't access satV reg" );
    displaySatV( currentSatV );
}


//
// 
//
static int
buildDeviceList( appResourcesPtr appRes, deviceTag* list )
{
    int index = 0;

    if ( appRes->xDeviceList == NULL ) return ERROR;

    // get string into a local buffer
    char* s = new char[ strlen(appRes->xDeviceList) + 1 ];
    strcpy( s, appRes->xDeviceList );

    // get the first label
    char* p = strtok( s, "," );
    SKIPWS( p );

    // process each of count possible values
    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	// set the index
	list[ x ].index = x;

	// assign the label
	list[ x ].label = p;

	// parse the resource file name
	if ( (p = strtok( NULL, "," )) == NULL )
	{
	    cerr << "buildDeviceList(), missing resource file" << endl;
	    return ERROR;
	}
	SKIPWS( p );

	// check for absolute or relative path
	if ( (*p == '/')
	     || ((p[0] == '.') && (p[1] == '/'))
	     || ((p[0] == '.') && (p[1] == '.') && (p[2] == '/')) )
	{
	    list[ x ].resFile = new char[ strlen( p ) + 1 ];
	    strcpy( list[ x ].resFile, p );
	}

	// no path, use default
	else
	{
	    list[ x ].resFile = new char[ strlen( p )
					  + strlen( RES_BASE ) + 1 ];
	    strcpy( list[ x ].resFile, RES_BASE );
	    strcat( list[ x ].resFile, p );
	}

	// get the next label
	if ( (p = strtok( NULL, "," )) == NULL ) break;
	SKIPWS( p );
    }

    // NULLify rest of list
    for ( ++x; x < NUM_DEVICES; ++x )
    {
	list[ x ].index = -1;
	list[ x ].label = NULL;
	list[ x ].resFile = NULL;
    }

    return OK;
}


//
// 
//
static int
buildFavorites( appResourcesPtr appRes, u_char* list )
{
    int index = 0;

    if ( appRes->xFavorites == NULL ) return ERROR;

    // get string into a local buffer
    char* s = new char[ strlen(appRes->xFavorites) + 1 ];
    strcpy( s, appRes->xFavorites );

    // get the first channel
    char* p = strtok( s, "," );
    SKIPWS( p );

    // process each of count possible values
    for ( int x = 0; x < MAX_CHANNEL; ++x )
    {
	// get the channel
	list[ x ] = atoi( p );

	// get the next channel
	if ( (p = strtok( NULL, "," )) == NULL ) break;
	SKIPWS( p );
    }

    // NULLify rest of list
    for ( favoriteCount = ++x; x < NUM_DEVICES; ++x )
    {
	list[ x ] = 0;
    }

    return OK;
}


//
// 
//
static int
loadDeviceResources( Widget w, deviceTag* list )
{
    XrmDatabase rdb = XtScreenDatabase( XtScreen( w ) );

    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].label == NULL ) break;

	// attempt to load rdb for device
	const char* s = list[ x ].resFile;
	if ( XrmCombineFileDatabase( s, &rdb, True /**False*/ ) == 0 )
	{
	    list[ x ].index = -1;
	    list[ x ].resFile = NULL;
	    cerr << "resFile NOT found: " << s << endl;
	}
    }

    return OK;
}


//
// 
//
static int
buildDeviceMenu( Widget menu, deviceTag* list, int& index )
{
    int x, c;
    Arg args[ 10 ];

    // compute total length
    for ( c = 0, x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].label == NULL ) break;
	c += (strlen( list[ x ].label ) + 2);
    }
    ++c;

    // build menu
    char* s = new char[ c ];
    s[ 0 ] = '\0';
    for ( x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].label == NULL ) break;	// end of list
	strcat( s, list[ x ].label );
	strcat( s, "\n" );
    }

    // find a valid default entry
    if ( index == -1 ) index = 0;
    if ( list[ index ].index == -1 )
    {
	for ( x = 0; x < NUM_DEVICES; ++x )
	{
	    if ( list[ x ].index != -1 )
	    {
		index = x;
		break;
	    }
	}
	if ( x == NUM_DEVICES ) return ERROR;
    }

    // set the menu string and default label
    int i = 0;
    XtSetArg( args[i], XtNlabel, list[ index ].label );  i++;
    XtSetArg( args[i], XtNmenu, s );  i++;
    XtSetValues( menu, args, i);

    return OK;
}


//
// 
//
static void
setActiveMenu( Widget menu, deviceTag* list )
{
    // search list for (unNULLified) entries
    long active = 0x00000000;
    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].index != -1 )		// valid entry
	{
	    active |= (1 << x);			// set bit in mask
	}
    }

    // find menu widget [created automatically by PullDown]
    Widget pw;
    Arg args[ 2 ];
    XtSetArg( args[ 0 ], XtNpopup, &pw );
    XtGetValues( menu, args, 1 );

    // set the 'active button list'
    XtSetArg( args[ 0 ], XtNactive, active );
    XtSetValues( pw, args, 1 );
}


static int
buildButtonBoard( Widget ButtonBoard, deviceTag* list )
{
    int count;
    int last;

    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	// end of list
	if ( list[ x ].label == NULL ) break;

	// handle sparse lists
	if ( list[ x ].index == -1 )
	{
	    devices[ x ].w = NULL;
	    devices[ x ].a = NULL;
	    continue;
	}
	/// cerr << "device: '" << list[ x ].label << "'" << endl;

	// clear the temp array
	for ( int xx = 0; xx < NUM_BUTTONS; ++xx )
	{
	    buttons[ xx ].w = NULL;
	    buttons[ xx ].a = NULL;
	}
	count = 0;
	last = -1;

	// create individual buttons
	const char* label = strrchr( list[ x ].resFile, '/' );
	if ( label == NULL ) label = list[ x ].resFile;
	else ++label;

	// build each defined button
	for ( xx = 0; xx < NUM_BUTTONS; ++xx )
	{
	    switch ( createButton( label,
				   ButtonBoard, xx, buttonCB ) )
	    {
	    case TRUE:
		++count;
		break;

	    case -1:
		last = xx;
		xx = NUM_BUTTONS;
		break;

	    case FALSE:
		break;
	    }
	}
	if ( last == -1 ) last = NUM_BUTTONS;

	// record count
	if ( (devices[ x ].c = count) == 0 )
	{
	    devices[ x ].w = NULL;
	    devices[ x ].a = NULL;
	    list[ x ].index = -1;
	    continue;
	}

	// build dense array of widgets
	int yy;
	devices[ x ].w = new Widget[ count ];
	for ( yy = 0, xx = 0; xx < last; ++xx )
	{
	    if ( buttons[ xx ].w )
	    {
		devices[ x ].w[ yy++ ] = buttons[ xx ].w;
	    }
	}

	if ( yy != count )
	{
	    cerr << "buildButtonBoard(), w != count" << endl;
	    cerr << "w: " << yy << "count: " << count << endl;
	    exit( 1 );
	}

	// build dense array of objects
	devices[ x ].a = new Action*[ count ];
	for ( yy = 0, xx = 0; xx < last; ++xx )
	{
	    if ( buttons[ xx ].a )
	    {
		// setup callback
		XtAddCallback( buttons[ xx ].w, XtNactivate,
			       buttonCB, (XtPointer)yy );
		// pack the object
		devices[ x ].a[ yy++ ] = buttons[ xx ].a;
	    }
	}

	if ( yy != count )
	{
	    cerr << "buildButtonBoard(), a != count" << endl;
	    cerr << "a: " << yy << "count: " << count << endl;
	    exit( 1 );
	}

	//
	char name[ 80 ];
	sprintf( name, "%s_size", label );

	// create an instance of it
	Widget w;
	w = XtVaCreateWidget( name, xfwfLabelWidgetClass, ButtonBoard, NULL );

	// get its label
	String message;
	Arg args[ 1 ];
	XtSetArg( args[ 0 ], XtNlabel, &message );
	XtGetValues( w, args, 1 );

	// delete label widget
	XtDestroyWidget( w );
	sscanf( message, "%dx%d", &devices[ x ].x, &devices[ x ].y );
	XtFree( message );
	/// cerr << "x: " << devices[x].x << ", y: " << devices[x].y << endl;
    }

    return OK;
}


//
// create push buttons
//
static int
createButton( const char* deviceName, Widget parent,
	       int number, XtCallbackProc callBack )
{
    // create a name for the message
    char name[ 256 ];
    sprintf( name, MESSAGE_RES_TMPL, deviceName, number );

#if defined( USE_XRMGETRESOURCE )

    would this be better???

#else  // USE_XRMGETRESOURCE

    // create an instance of it
    Widget w;
    w = XtVaCreateWidget( name, xfwfLabelWidgetClass, parent, NULL );

    // get its label
    String message;
    Arg args[ 1 ];
    XtSetArg( args[ 0 ], XtNlabel, &message );
    XtGetValues( w, args, 1 );

    // delete label widget
    XtDestroyWidget( w );

#endif // USE_XRMGETRESOURCE

    // test it
    if ( message )
    {
	Action* action = addMessage( message );
	XtFree( message );

	// found last message
	if ( action == NULL ) return (-1);

	// record buttons action
	buttons[ number ].a = action;

	// create the actual button
	sprintf( name, BUTTON_RES_TMPL, deviceName, number );
	w = XtVaCreateWidget( name, xfwfButtonWidgetClass, parent, NULL );

        // funky...
        if ( widgetPointer )
        {
            *widgetPointer = w;
            widgetPointer = NULL;
        }

	// record widget
	buttons[ number ].w = w;
	return TRUE;
    }

    else
    {
	buttons[ number ].w = NULL;
	buttons[ number ].a = NULL;
	return FALSE;
    }
}


//
// options menu callback
//
static void
optionsCB( Widget w, XtPointer clientData, XtPointer callData )
{
    XfwfTextMenuData& data = *(XfwfTextMenuData*)callData;

    switch ( data.n )
    {
    case 0:
	exit(0);

    default:
	break;
    }
}


//
// device menu callback
//
static void
deviceCB( Widget w, XtPointer clientData, XtPointer callData )
{
    XfwfTextMenuData& data = *(XfwfTextMenuData*)callData;

    // check for invalid device
    if ( deviceList[ data.n ].index == -1 )
    {
	return;
    }

    // unmap the current device buttons
    XtUnmanageChildren( devices[ xDevice ].w, devices[ xDevice ].c );

    // get the new device type
    xDevice = deviceList[ data.n ].index;

    // change size
    Arg args[ 2 ];
    int i = 0;
    XtSetArg( args[i], XtNwidth, devices[ xDevice ].x );  i++;
    XtSetArg( args[i], XtNheight, devices[ xDevice ].y );  i++;
    XtSetValues( topLevel, args, i );

    // map the new device buttons
    XtManageChildren( devices[ xDevice ].w, devices[ xDevice ].c );
}


//
// button pressed callback
//
static void
buttonCB( Widget w, XtPointer clientData, XtPointer callData )
{
    if ( devices[ xDevice ].a[ (int)clientData ] )
	devices[ xDevice ].a[ (int)clientData ]->action( w );
}


/////////////////////////////// tuner functions ///////////////////////////////

//
// 
//
static int
getSource( void )
{
    int x;

    if ( ioctl( video, METEORGINPUT, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    switch ( x )
    {
    case METEOR_DEV0:	// the RCA jack
        return 1;
    case METEOR_DEV1:	// the tuner
        return 2;
    case METEOR_DEV2:	// the SVHS input
        return 3;
    default:
        return ERROR;
    }
}

//
// 
//
static int
setSource( Widget w, int source )
{
    int x;

    if ( (source < 1) || (source > 3) )
      return ERROR;

    x = sources[ source-1 ].source;
    if ( ioctl( video, METEORSINPUT, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    currentSource = source;
    XtVaSetValues( w, XtNlabel, sources[ source-1 ].name, NULL );

    return OK;
}

//
// 
//
static int
toggleSource( Widget w )
{
    int x;

    switch ( currentSource )
    {
    case 1:
        x = 2;
        break;
    case 2:
        x = 3;
        break;
    case 3:
        x = 1;
        break;
    default:
        return ERROR;
    }

    return setSource( w, x );
}


//
// #define TUNERTYPE_NABCST	1
// #define TUNERTYPE_CABLEIRC	2
// #define TUNERTYPE_CABLEHRC	3
// #define TUNERTYPE_WEUROPE	4
//


//
// 
//
static int
getTunerType( void )
{
    int x;

    if ( ioctl( video, TVTUNER_GETTYPE, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    return x;
}

//
// 
//
static int
setTunerType( Widget w, int tunerType )
{
    if ( ioctl( video, TVTUNER_SETTYPE, &tunerType ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    currentTunerType = tunerType;
    XtVaSetValues( w, XtNlabel, tunerTypes[ tunerType-1 ].name, NULL );

    return OK;
}

//
// 
//
static int
toggletuner( Widget w )
{
    int x;

    // FIXME: do this from a resource list
    switch ( currentTunerType )
    {
    case 1:
        x = 2;
        break;
    case 2:
        x = 1;
        break;
    case 3:
    case 4:
    default:
        return ERROR;
    }

    return setTunerType( w, x );
}


//
// 
//
static int
updateChannel( int channel )
{
    char label[ 128 ];

    previousChannel = currentChannel;
    currentChannel = channel;

#if 0
    XfwfLabelSetLabel( chnlDisplay, label );
#else
    sprintf( label, "Channel %d", channel );
    XtVaSetValues( chnlDisplay, XtNlabel, label, NULL );
#endif
}


//
// 
//
static int
setChannel( Widget w, char* arg )
{
    int x;

    x = atoi( arg );

    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return OK;
}


//
// 
//
static int
upChannel( Widget w, void )
{
    int x;

    if ( currentChannel )
    {
        for ( int y = 0; y < favoriteCount; ++y )
        {
            if ( favoriteChannels[ y ] > currentChannel )	// found
            {
                x = y;
                break;
            }
        }

        if ( y == favoriteCount )				// wrapped
        {
            x = 0;
        }
    }
    else
    {
        x = 0;
    }

    if ( !favoriteChannels[ x ] ) return ERROR;

    x = favoriteChannels[ x ];
    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return OK;
}


//
// 
//
static int
downChannel( Widget w, void )
{
    int x;

    if ( currentChannel )
    {
        for ( int y = favoriteCount - 1; y > -1 ; --y )
        {
            if ( favoriteChannels[ y ] < currentChannel )	// found
            {
                x = y;
                break;
            }
        }

        if ( y == -1 )						// wrapped
        {
            x = favoriteCount - 1;
        }
    }
    else
    {
        x = favoriteCount - 1;
    }

    if ( !favoriteChannels[ x ] ) return ERROR;

    x = favoriteChannels[ x ];
    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return OK;
}


//
// 
//
static int
getChannel( Widget w )
{
    int x;

    if ( ioctl( video, TVTUNER_GETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return x;
}


//
// 
//
static int
getStatus( Widget w )
{
    int x;

    if ( ioctl( video, TVTUNER_GETSTATUS, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    fprintf( stderr, "0x%02x\n", x );

    return x;
}


//
// 
//
static int
setPrevious( Widget w, void )
{
    int x;

    if ( !previousChannel ) return 0;

    x = previousChannel;
    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return x;
}


static void
enterInputLabel( Widget w, void )
{
    char label[ 128 ];

#if 0
    XfwfLabelSetLabel( enterInputWidget, label );
#else
    if ( inputChannel )
        sprintf( label, "Enter\n%d", inputChannel );
    else
        strcpy( label, "Enter" );

    XtVaSetValues( enterInputWidget, XtNlabel, label, NULL );
#endif
}

//
// 
//
static int
enterInput( Widget w, void )
{
    int x;

    if ( !inputChannel ) return 0;

    x = inputChannel;
    inputChannel = inputDigits = 0;

    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return x;
}

//
// 
//
static int
addInput( Widget w, char* input )
{
    if ( inputDigits >= 3 )
    {
        inputChannel = 0;
        inputDigits = 0;
    }

    ++inputDigits;
    inputChannel = (inputChannel * 10) + atoi( input );

    enterInputLabel( w );

    return OK;
}

//
// 
//
static int
clearInput( Widget w )
{
    inputChannel = 0;

    enterInputLabel( w );

    return OK;
}

///////////////////////////////////////////////////////////////////////////////
// functions to control the picture

#if 0
static int	percentify( int value, int limit );
//
// 
//
static int
percentify( int value, int limit )
{
    int x = (int)(((double)value / (double)limit) * 100.0);
    if ( x < 0 )
      return 0;
    if ( x > 100 )
      return 100;
    return x;
}
#endif


//
// 
//
static int
normalize( int* value,
           int minValue, int maxValue,
           double range, int steps,
           int minReg, int maxReg )
{
    int		x;

    // keep the LOGICAL value between LOGICAL min and max
    if ( *value > maxValue )
      *value = maxValue;
    else if ( *value < minValue )
      *value = minValue;

    // calculate register value needed for value
    x = (int)rint( (double)(*value) / (range/(double)steps) );

    // keep the register value between register min and max
    if ( x > maxReg )
      x = maxReg;
    else if ( x < minReg )
      x = minReg;

    return x;
}


//
// 
//
static int
getHue( Widget w )
{
    int		value;

    if ( ioctl( video, BT848_GHUE, &value ) < 0 )
      return -1;

    return (int)rint( (BT848_HUERANGE / BT848_HUESTEPS) * (double)value );
}


//
// 
//
static int
bumpVolume( Widget w, char* arg )
{
    int		value;
    char	s[ 512 ];

    value = atoi( arg ) + currentVolume;
    if ( value < 0 )
      value = 0;
    else if ( value > 100 )
      value = 100;

    snprintf( s, 512, "%s %d > /dev/null", mixerCommand, value );
    system( s );

    return ( currentVolume = value );
}


//
// 
//
static int
bumpHue( Widget w, char* arg )
{
    int		x;
    int		value;

    /* new hue in integer degrees */
    value = atoi( arg ) + currentHue;
    x = normalize( &value,
                   BT848_HUEMIN, BT848_HUEMAX,
                   BT848_HUERANGE, BT848_HUESTEPS,
                   BT848_HUEREGMIN, BT848_HUEREGMAX );

    if ( ioctl( video, BT848_SHUE, &x ) < 0 )
      return -1;

    displayHue( currentHue = value );

    return currentHue;
}


static void
displayHue( int value )
{
    char label[ 128 ];
    sprintf( label, "Hue %+d'", value );
    XtVaSetValues( hueDisplay, XtNlabel, label, NULL );
}


//
// 
//
static int
getBright( Widget w )
{
    int	value;

    if ( ioctl( video, BT848_GBRIG, &value ) < 0 )
      return -1;

    return (int)rint( (BT848_BRIGHTRANGE / BT848_BRIGHTSTEPS)
                      * (double)value );
}


//
// 
//
static int
bumpBright( Widget w, char* arg )
{
    int		x;
    int		value;

    /* new brightness in integer % */
    value = atoi( arg ) + currentBright;
    x = normalize( &value,
                   BT848_BRIGHTMIN, BT848_BRIGHTMAX,
                   BT848_BRIGHTRANGE, BT848_BRIGHTSTEPS,
                   BT848_BRIGHTREGMIN, BT848_BRIGHTREGMAX );

    if ( ioctl( video, BT848_SBRIG, &x ) < 0 )
      return -1;

    displayBright( currentBright = value );

    return currentBright;
}


static void
displayBright( int value )
{
    char label[ 128 ];
    sprintf( label, "Brightness %+d%%", value );
    XtVaSetValues( brightDisplay, XtNlabel, label, NULL );
}


//
// 
//
static int
getContrast( Widget w )
{
    int	value;

    if ( ioctl( video, BT848_GCONT, &value ) < 0 )
      return -1;

    return (int)rint( (BT848_CONTRASTRANGE / BT848_CONTRASTSTEPS)
                      * (double)value );
}


//
// 
//
static int
bumpContrast( Widget w, char* arg )
{
    int		x;
    int		value;

    value = atoi( arg ) + currentContrast;
    x = normalize( &value,
                   BT848_CONTRASTMIN, BT848_CONTRASTMAX,
                   BT848_CONTRASTRANGE, BT848_CONTRASTSTEPS,
                   BT848_CONTRASTREGMIN, BT848_CONTRASTREGMAX );

    if ( ioctl( video, BT848_SCONT, &x ) < 0 )
      return -1;

    displayContrast( currentContrast = value );

    return currentContrast;
}


static void
displayContrast( int value )
{
    char label[ 128 ];
    sprintf( label, "Contrast %+d%%", value - BT848_CONTRASTCENTER );
    XtVaSetValues( contrastDisplay, XtNlabel, label, NULL );
}


//
// 
//
static int
getChroma( Widget w )
{
    int	value;

    if ( ioctl( video, BT848_GCSAT, &value ) < 0 )
      return -1;

    return (int)rint( (BT848_CHROMARANGE / BT848_CHROMASTEPS)
                      * (double)value );
}


//
// 
//
static int
bumpChroma( Widget w, char* arg )
{
    int		x;
    int		value;

    value = atoi( arg ) + currentChroma;

    // check LOGICAL range
    if ( value > BT848_SATVMAX )	// the larger of U and V
      value = BT848_SATVMAX;
    else if ( value < BT848_SATVMIN )	// both have same minimum
      value = BT848_SATVMIN;

    // adjust the 2 components
    bumpSatU( w, arg );
    bumpSatV( w, arg );

    displayChroma( currentChroma = value );

    return currentChroma;
}


static void
displayChroma( int value )
{
    char label[ 128 ];
    sprintf( label, "Chroma %+d%%", value - BT848_CHROMACENTER );
    XtVaSetValues( chromaDisplay, XtNlabel, label, NULL );
}


//
// 
//
static int
getSatU( Widget w )
{
    int	value;

    if ( ioctl( video, BT848_GUSAT, &value ) < 0 )
      return -1;

    return (int)rint( (BT848_SATURANGE / BT848_SATUSTEPS) * (double)value );
}


#if 0
//
// 
//
static int
rawSatU( void )
{
    int	value;

    if ( ioctl( video, BT848_GUSAT, &value ) < 0 )
      return -1;

    return value;
}
#endif


//
// 
//
static int
bumpSatU( Widget w, char* arg )
{
    int		x;
    int		value;

    value = atoi( arg ) + currentSatU;
    x = normalize( &value,
                   BT848_SATUMIN, BT848_SATUMAX,
                   BT848_SATURANGE, BT848_SATUSTEPS,
                   BT848_SATUREGMIN, BT848_SATUREGMAX );

    if ( ioctl( video, BT848_SUSAT, &x ) < 0 )
      return -1;

    displaySatU( currentSatU = value );

    return currentSatU;
}


static void
displaySatU( int value )
{
    char label[ 128 ];
#if 0
    sprintf( label, "SatU %+d%% (%d)", value - BT848_SATUCENTER, rawSatU() );
#else
    sprintf( label, "SatU %+d%%", value - BT848_SATUCENTER );
#endif
    XtVaSetValues( satUDisplay, XtNlabel, label, NULL );
}


//
// 
//
static int
getSatV( Widget w )
{
    int	value;

    if ( ioctl( video, BT848_GVSAT, &value ) < 0 )
      return -1;

    return (int)rint( (BT848_SATVRANGE/BT848_SATVSTEPS) * (double)value );
}


#if 0
//
// 
//
static int
rawSatV( void )
{
    int	value;

    if ( ioctl( video, BT848_GVSAT, &value ) < 0 )
      return -1;

    return value;
}
#endif


//
// 
//
static int
bumpSatV( Widget w, char* arg )
{
    int		x;
    int		value;

    value = atoi( arg ) + currentSatV;
    x = normalize( &value,
                   BT848_SATVMIN, BT848_SATVMAX,
                   BT848_SATVRANGE, BT848_SATVSTEPS,
                   BT848_SATVREGMIN, BT848_SATVREGMAX );

    if ( ioctl( video, BT848_SVSAT, &x ) < 0 )
      return -1;

    displaySatV( currentSatV = value );

    return currentSatV;
}


static void
displaySatV( int value )
{
    char label[ 128 ];
#if 0
    sprintf( label, "SatV %+d%% (%d)", value - BT848_SATVCENTER, rawSatV() );
#else
    sprintf( label, "SatV %+d%%", value - BT848_SATVCENTER );
#endif
    XtVaSetValues( satVDisplay, XtNlabel, label, NULL );
}


//
// 
//
static int
toggleMute( Widget w )
{
    int x;

    switch ( currentMute )
    {
    case 1:
        x = AUDIO_UNMUTE;
        if ( ioctl( video, BT848_SAUDIO, &x ) < 0 )
          return -1;
        currentMute = 0;
        break;

    case 0:
        x = AUDIO_MUTE;
        if ( ioctl( video, BT848_SAUDIO, &x ) < 0 )
          return -1;
        currentMute = 1;
        break;

    default:
        return ERROR;
    }

    return 0;
}


//
// 
//
static int
noop( Widget w, void )
{
}

static Action* NaddMessage( char* msg );


#define CT_END		1
#define CT_SYS		2
#define CT_FUNC		3

typedef		int(*xfunc_t)(Widget w);	// FIXME

struct COMMAND {
	char*	name;
	int	type;
        func_t	func;
        Widget* wp;
} commands[] = {
  { "ADD_INPUT",	CT_FUNC, (func_t)addInput,	NULL },
  { "BUMP_BRIGHT",	CT_FUNC, (func_t)bumpBright,	NULL },
  { "BUMP_CHROMA",	CT_FUNC, (func_t)bumpChroma,	NULL },
  { "BUMP_CONTRAST",	CT_FUNC, (func_t)bumpContrast,	NULL },
  { "BUMP_HUE",		CT_FUNC, (func_t)bumpHue,	NULL },
  { "BUMP_SATU",	CT_FUNC, (func_t)bumpSatU,	NULL },
  { "BUMP_SATV",	CT_FUNC, (func_t)bumpSatV,	NULL },
  { "BUMP_VOLUME",	CT_FUNC, (func_t)bumpVolume,	NULL },
  { "CLEAR_INPUT",	CT_FUNC, (func_t)clearInput,	NULL },
  { "DISPLAY_BRIGHT",	CT_FUNC, (func_t)noop,		&brightDisplay },
  { "DISPLAY_CHNL",	CT_FUNC, (func_t)noop,		&chnlDisplay },
  { "DISPLAY_CHROMA",	CT_FUNC, (func_t)noop,		&chromaDisplay },
  { "DISPLAY_CONTRAST",	CT_FUNC, (func_t)noop,		&contrastDisplay },
  { "DISPLAY_HUE",	CT_FUNC, (func_t)noop,		&hueDisplay },
  { "DISPLAY_SATU",	CT_FUNC, (func_t)noop,		&satUDisplay },
  { "DISPLAY_SATV",	CT_FUNC, (func_t)noop,		&satVDisplay },
  { "DN_CHNL",		CT_FUNC, (func_t)downChannel,	NULL },
  { "ENTER_INPUT",	CT_FUNC, (func_t)enterInput,	&enterInputWidget },
  { "GET_BRIGHT",	CT_FUNC, (func_t)getBright,	NULL },
  { "GET_CHNL",		CT_FUNC, (func_t)getChannel,	NULL },
  { "GET_CHROMA",	CT_FUNC, (func_t)getChroma,	NULL },
  { "GET_CONTRAST",	CT_FUNC, (func_t)getContrast,	NULL },
  { "GET_HUE",		CT_FUNC, (func_t)getHue,	NULL },
  { "GET_SATU",		CT_FUNC, (func_t)getSatU,	NULL },
  { "GET_SATV",		CT_FUNC, (func_t)getSatV,	NULL },
  { "GET_STATUS",	CT_FUNC, (func_t)getStatus,	NULL },
  { "MSG_END",		CT_END,  NULL,			NULL },
  { "NOOP",		CT_FUNC, (func_t)noop,		NULL },
  { "PREVIOUS",		CT_FUNC, (func_t)setPrevious,	NULL },
  { "SET_CHNL",		CT_FUNC, (func_t)setChannel,	NULL },
  { "SYSTEM",		CT_SYS,  NULL,			NULL },
  { "TOGGLE_MUTE",	CT_FUNC, (func_t)toggleMute,	NULL },
  { "TOGGLE_SOURCE",	CT_FUNC, (func_t)toggleSource,	&sourceTglW },
  { "TOGGLE_TUNERTYPE",	CT_FUNC, (func_t)toggletuner,	&tunerTglW },
  { "UP_CHNL",		CT_FUNC, (func_t)upChannel,	NULL  }
};


static int
compare( const void* p1, const void* p2 )
{
    return strcmp( ((const struct COMMAND*)p1)->name,
                   ((const struct COMMAND*)p2)->name );
}


//
// create an action object for the button
//
static Action*
addMessage( char* msg )
{
    Action*		object;
    char*		stringp;
    struct COMMAND*	ptr;
    struct COMMAND	cmd;

    // make a working copy
    char* args = new char[ strlen( msg ) + 1 ];
    strcpy( args, msg );

    // break string up into address & message
    stringp = args;
    strsep( &stringp, " 	" );

    // lookup the function
    cmd.name = args;
    ptr = (struct COMMAND*) bsearch( &cmd, commands,
                                     sizeof(commands) / sizeof(struct COMMAND),
                                     sizeof(struct COMMAND), compare );
    
    // not found...
    if ( ptr == NULL )
    {
        fprintf( stderr, "NOT found: '%s'\n", args );
        exit( 1 );
    }

    // install it as appropriate
    switch ( ptr->type )
    {
    case CT_END:
	return NULL;
        break;
    case CT_SYS:
	object = new System( stringp );
        break;
    case CT_FUNC:
        if ( stringp )
          object = new Function( (func_t)ptr->func, stringp );
        else
          object = new Function( (xfunc_t)ptr->func );
        if ( ptr->wp )
          widgetPointer = ptr->wp;
        break;
    default:
	object = new Function( noop );
    }

    // return memory
    delete args;

    return object;
}
